<?php

namespace Drupal\redirects_fixer;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Component\Utility\Html;
use Drupal\Component\Utility\Unicode;
use Drupal\Component\Utility\UrlHelper;
use Drupal\Core\Cache\Cache;
use Drupal\Core\DependencyInjection\DependencySerializationTrait;
use Drupal\Core\Logger\LoggerChannelTrait;
use Drupal\Core\Messenger\MessengerTrait;
use Drupal\Core\State\StateInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Url;
use Drupal\linkit\Utility\LinkitHelper;
use Drupal\redirects_fixer\Entity\RedirectsFixerLogEntity;

/**
 * Batch manager.
 */
class BatchManager {

  use StringTranslationTrait;
  use DependencySerializationTrait;
  use MessengerTrait;
  use LoggerChannelTrait;

  const STATE_KEY_ENTITY_TYPE = 'links_redirect_fixer_entity_type';
  const STATE_KEY_URLS = 'links_redirect_fixer_urls';
  const STATE_KEY_COUNT_UPDATED_LINKS = 'links_redirect_fixer_count';
  const STATE_KEY_IDS = 'links_redirect_fixer_ids';

  const FIELDS_WITH_LINKS_MAP = [
    'block_content' => 'body',
    'node' => 'body',
  ];

  protected $total;
  protected $logs;
  protected $changedValueEntities;
  protected $linksFinder;
  protected $linksManager;
  protected $state;
  protected $baseUrl;
  protected $time;
  protected string $baseDomain;
  private \Drupal\Core\Config\ImmutableConfig $config;

  /**
   * Constructs a BatchManager object.
   */
  public function __construct(
    LinksFinder $links_finder,
    LinksManager $links_manager,
    StateInterface $state,
    TimeInterface $time,
  ) {
    $this->linksFinder = $links_finder;
    $this->linksManager = $links_manager;
    $this->state = $state;
    $this->time = $time;

    // Use prod full path for preparing links to request.
    $config = \Drupal::configFactory()->get('redirects_fixer.settings');
    $site_domain = $config->get('site-domain');
    $this->config = $config;
    $this->baseDomain = $site_domain;
    $this->baseUrl = 'https://' . $site_domain;
  }

  /**
   * Prepare butch data for a process.
   */
  public function createBatchData(string $entity_type_id): array {
    $this->state->delete(static::STATE_KEY_URLS);
    $this->state->delete(static::STATE_KEY_COUNT_UPDATED_LINKS);
    $this->state->delete(static::STATE_KEY_IDS);
    $this->state->set(static::STATE_KEY_ENTITY_TYPE, $entity_type_id);

    $total_entities = $this->getTotal($entity_type_id);

    $batch = [
      'title' => $this->t("Scan body fields for @entity_type", ['@entity_type' => Unicode::ucfirst($entity_type_id)]),
      'init_message' => $this->t('Start update @count entities', ['@count' => $total_entities]),
      'finished' => [$this, 'batchFinish'],
      'operations' => [],
    ];

    $limit = $entity_type_id === 'block_content'
      ? $this->config->get('items_per_batch_block_content')
      : $this->config->get('items_per_batch_node');
    // For debug purposed set the limit to same number of elements as defined
    // in \Drupal\redirects_fixer\LinksFinder::DEBUG_NIDS.
    if ($this->linksFinder::DEBUG) {
      $limit = count($this->linksFinder::DEBUG_NIDS);
    }

    $iterations = (int) ($total_entities / $limit);
    $operation_start_timestamp = $this->time->getRequestTime();
    for ($iteration = 0; $iteration < $iterations; $iteration++) {
      $batch['operations'][] = [
        [$this, 'operationCallback'],
        [$iteration, $entity_type_id, $limit, $operation_start_timestamp],
      ];
    }

    return $batch;
  }

  /**
   * Batch callback.
   */
  public function operationCallback($iteration, $entity_type_id, $limit, $operation_start_timestamp, array &$context): void {
    $table = $entity_type_id . '__' . static::FIELDS_WITH_LINKS_MAP[$entity_type_id];
    $table_revision = $entity_type_id . '_revision__' . static::FIELDS_WITH_LINKS_MAP[$entity_type_id];
    $value_column_name = static::FIELDS_WITH_LINKS_MAP[$entity_type_id] . '_value';

    $links_fields = $this->linksFinder
      ->getLinksFields($table, $value_column_name, $iteration * $limit, $limit);

    $links_data = $this->linksFinder->prepareLinksData($links_fields);

    $urls_to_process = [];
    $links_to_process = [];
    $urls_to_process_for_logging = [];
    foreach ($links_data as $entity_id => $data) {
      /** @var \DOMElement $link_element */
      foreach ($data['list'] as $link_element) {
        if ($this->isApplicable($link_element)) {
          $urls_to_process[] = $this->prepareUrlToProcess($link_element, $entity_type_id, $entity_id);
          $links_to_process[] = ['id' => $entity_id, 'link' => $link_element];
          try {
            $urls_to_process_for_logging[] = $link_element->getAttribute('href');
          }
          catch (\Exception $exception) {
            \Drupal::logger('redirects_fixer')->error($exception->getMessage()
              . ' ' . $link_element->ownerDocument->saveXML($link_element));
            $context['message'] = $this->t('Error: ') . $exception->getMessage() . $link_element->ownerDocument->saveXML($link_element);
          }
        }
      }
    }

    $context['message'] = new TranslatableMarkup('Now processing the links: <br> @links', ['@links' => new TranslatableMarkup(implode(' <br>', $urls_to_process_for_logging))]);

    $url_links_replace_map = $this->processUrls(array_unique($urls_to_process));

    foreach ($links_to_process as $link_data) {
      $link_element = $link_data['link'];
      if (!empty($url_links_replace_map[$this->prepareUrlToProcess($link_element, $entity_type_id, $link_data['id'])])) {
        $this->processedUrlToLinkElement(
          $link_element,
          $url_links_replace_map[$this->prepareUrlToProcess($link_element, $entity_type_id, $link_data['id'])],
          $link_data['id'],
          $entity_type_id,
          $operation_start_timestamp,
        );
      }
    }

    if (!$this->changedValueEntities) {
      return;
    }

    $this->changedValueEntities = array_unique($this->changedValueEntities);

    $new_links_fields = [];
    foreach ($links_data as $id => $data) {
      if (in_array($id, $this->changedValueEntities)) {
        $new_links_fields[$id] = Html::serialize($data['document']);
      }
    }

    $this->linksFinder->updateLinksData($table, $value_column_name, $new_links_fields, $table_revision);

    // Logs are not saving.
    $this->saveLogs();
    $this->state->set(
      static::STATE_KEY_IDS,
      array_merge($this->state->get(static::STATE_KEY_IDS, []), $this->changedValueEntities),
    );

  }

  /**
   * Check that url is applicable.
   */
  protected function isApplicable(\DOMElement $link_element): bool {
    $url = $link_element->getAttribute('href');
    try {
      if (filter_var($url, FILTER_VALIDATE_EMAIL) || str_starts_with($url, 'mailto:')) {
        return FALSE;
      }

      $valid = UrlHelper::isValid($url);
      $internal = $valid && $this->isInternalUrl($url);
      if (!$valid || !$internal) {
        return FALSE;
      }

      // It is possible that external links like 'pathguy.com', 'keep.google.com' exists without a protocol.
      // So it is not applicable to test them for internal redirects.
      if ($internal) {
        preg_match("/^([a-zA-Z0-9-_]+(\.[a-zA-Z0-9-_]+)+)/", $url, $matches);
        if (!empty($matches[0])) {
          if ($matches[0] !== $this->baseDomain) {
            return FALSE;
          }
        }
      }

      $path = parse_url($url, PHP_URL_PATH);
      if (empty($path) || $path === '/') {
        return FALSE;
      }

      $is_linkit_file = $this->isLinkit($link_element) && $link_element->getAttribute('data-entity-substitution') === 'file';
      if ($is_linkit_file || $this->isFile($url)) {
        return FALSE;
      }

      return TRUE;
    }
    catch (\Exception $e) {
      return FALSE;
    }
  }

  /**
   * Check that url fo file.
   */
  protected function isFile(string $url): bool {
    $file_extensions = [
      'jpg', 'jpeg', 'png', 'gif', 'bmp', 'webp', // Image files
      'pdf', 'doc', 'docx', 'xls', 'xlsx', 'ppt', 'pptx', // Document files
      'txt', 'csv', 'md', // Text files
      'zip', 'rar', '7z', 'tar', 'gz', // Archive files
      'mp3', 'wav', 'ogg', // Audio files
      'mp4', 'avi', 'mov', 'mkv', // Video files
    ];
    $path = parse_url($url, PHP_URL_PATH);
    $extension = pathinfo($path, PATHINFO_EXTENSION);
    return in_array(strtolower($extension), $file_extensions);
  }

  /**
   * Prepare url to process.
   */
  protected function prepareUrlToProcess(\DOMElement $link_element, string $entity_type_id, int|string $entity_id): string {

    $url = $link_element->getAttribute('href');
    // In case we have broken links without protocol, like "<a href="{$this->baseDomain}/path">Title</a>".
    // Fix them here.
    if (str_starts_with($url, $this->baseDomain)) {
      return  'https://' . $url;
    }

    if (!str_starts_with($url, 'http') && !str_starts_with($url, '/')) {
      $url = '//' . $url;
    }

    $path = parse_url($url, PHP_URL_PATH) ?? '';
    if (empty($path)) {
      \Drupal::logger('redirects_fixer__prepareUrlToProcess')
        ->warning( 'Cannot parse url: @url, entity_type_id: @entity_type_id, entity_id: @entity_id', [
          '@url' => $url,
          '@entity_type_id' => $entity_type_id,
          '@entity_id' => $entity_id,
          ]
        );
    }

    if (!str_starts_with($path, '/')) {
      $path = '/' . $path;
    }

    return $this->baseUrl . $path;
  }

  /**
   * Process urls to determinate redirects.
   */
  protected function processUrls(array $urls_to_process): array {
    $processed = [];
    $cached_processed = $this->state->get(static::STATE_KEY_URLS, []);
    foreach (array_intersect(array_keys($cached_processed), $urls_to_process) as $cached_url) {
      $processed[$cached_url] = $cached_processed[$cached_url];
    }

    $result = [];
    $to_process = array_diff($urls_to_process, array_keys($cached_processed));
    if (count($to_process) > 20) {
      $chunks = array_chunk($to_process, 20);
      foreach ($chunks as $chunk) {
        $result = array_merge(
          $result,
          $this->linksManager->urlsDestination($chunk),
        );
      }
    }
    else {
      $result = $this->linksManager->urlsDestination($to_process);
    }

    $this->state->set(static::STATE_KEY_URLS, array_merge($cached_processed, $result));
    return array_merge($processed, $result);
  }

  /**
   * Check that link is linkit.
   */
  protected function isLinkit(\DOMElement $link_element): bool {
    // @see \Drupal\linkit\Plugin\Filter\LinkitFilter::process()
    return $link_element->hasAttribute('data-entity-type')
      && $link_element->hasAttribute('data-entity-uuid');
  }

  /**
   * Check that url is internal.
   */
  protected function isInternalUrl(string $url): bool {
    if (!UrlHelper::isExternal($url)) {
      return TRUE;
    }

    $main_domain_pattern = $this->baseDomain;
    $parsed_url = parse_url($url);
    $host = $parsed_url['host'] ?? $parsed_url['path'];
    $normalized_host = preg_replace('/^www\./', '', $host);
    return $normalized_host === $main_domain_pattern;
  }

  /**
   * Set processed urls to link element.
   *
   * @todo Add clear description, list parameters.
   */
  protected function processedUrlToLinkElement(\DOMElement $link_element, string $new_url, int $entity_id, string $entity_type, int $operation_start_timestamp): void {
    $original_url = $link_element->getAttribute('href');
    if (UrlHelper::isExternal($new_url)) {
      $new_url = parse_url($new_url, PHP_URL_PATH) ?? '/';
    }

    $new_original_path = LinkitHelper::getPathByAlias($new_url);
    if ($original_url === $new_original_path) {
      return;
    }

    $has_canonical = $new_original_path !== $new_url;
    $is_linkit = $this->isLinkit($link_element);

    // Preserve original query parameters and fragment if any.
    $new_original_path .= LinkitHelper::getQueryAndFragment($original_url);

    if ($has_canonical) {
      $get_entity = function () use ($new_url) {
        return LinkitHelper::getEntityFromUri(LinkitHelper::uriFromUserInput($new_url));
      };
      $this->linksManager->updateLinkitLink($link_element, $new_original_path, $get_entity());
    }
    else {
      $this->linksManager->updateLink($link_element, $new_original_path);
    }

    $this->changedValueEntities[] = $entity_id;

    $this->logs[] = [
      'entity_id'   => $entity_id,
      'entity_type' => $entity_type,
      'time'        => $operation_start_timestamp,
      'old_type'    => $is_linkit ? 'linkit' : 'link',
      'new_type'    => $this->isLinkit($link_element) ? 'linkit' : 'link',
      'old_url'     => $original_url,
      'new_url'     => $new_original_path,
    ];

    $count = $this->state->get(static::STATE_KEY_COUNT_UPDATED_LINKS, 0);
    $this->state->set(static::STATE_KEY_COUNT_UPDATED_LINKS, ++$count);
  }

  /**
   * Get total rows (entities) for process.
   */
  public function getTotal(string $entity_type_id): int {
    if ($this->total) {
      return $this->total;
    }

    $table = $entity_type_id . '__' . static::FIELDS_WITH_LINKS_MAP[$entity_type_id];
    $field = static::FIELDS_WITH_LINKS_MAP[$entity_type_id] . '_value';

    return $this->total = $this->linksFinder->totalCountFields($table, $field);
  }

  /**
   * Save static stored logs.
   */
  protected function saveLogs(): void {
    foreach ($this->logs as $item) {
      RedirectsFixerLogEntity::create($item)->save();
    }
  }

  /**
   * Invalidate cache of updated entities.
   */
  protected function invalidateCache(string $entity_type_id, array $ids): void {
    $tags =[];
    foreach ($ids as $id) {
      $tags[] = $entity_type_id . ':' . $id;
    }
    Cache::invalidateTags($tags);
  }

  /**
   * Finish batch callback.
   */
  public function batchFinish(): void {
    $ids = $this->state->get(static::STATE_KEY_IDS, []);
    if ($ids) {
      $this->invalidateCache($this->state->get(static::STATE_KEY_ENTITY_TYPE), $ids);
    }

    $count = $this->state->get(static::STATE_KEY_COUNT_UPDATED_LINKS, 0);
    $this->getLogger('redirect_fixer')->info("Updated {$count} links.");
    $db_log_url = Url::fromUserInput('/admin/reports/dblog')->toString();
    $this->messenger()->addMessage(new TranslatableMarkup("Updated $count links. See logs for additional errors: <a href=\"$db_log_url\">Error log</a>"));

    $this->state->delete(static::STATE_KEY_ENTITY_TYPE);
    $this->state->delete(static::STATE_KEY_URLS);
    $this->state->delete(static::STATE_KEY_COUNT_UPDATED_LINKS);
    $this->state->delete(static::STATE_KEY_IDS);
  }

}
