<?php

namespace Drupal\redirects_fixer\Controller;

use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Url;
use Drupal\redirects_fixer\BatchManager;
use Drupal\redirects_fixer\Form\SettingsForm;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RequestStack;

/**
 * Controller Fix Links.
 */
class FixLinksController extends ControllerBase {

  /**
   * Batch manager.
   *
   * @var \Drupal\redirects_fixer\BatchManager
   */
  protected BatchManager $batchManager;

  /**
   * The request stack.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected RequestStack $requestStack;

  /**
   * Constructs a FixLinksController object.
   */
  public function __construct(BatchManager $batch_manager, RequestStack $request_stack) {
    $this->batchManager = $batch_manager;
    $this->requestStack = $request_stack;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('redirects_fixer.batch_manager'),
      $container->get('request_stack'),
    );
  }

  /**
   * Starts the page by building the necessary elements.
   *
   * @return array
   *   An array containing the page elements.
   */
  public function startPage() {
    $build['info'] = [
      '#markup' => $this->t('This tool is designed to iterate through all nodes and blocks on the site, identify links in the Body field, and replace them with redirect destinations if the links point to non-final pages. <br>'),
      '#theme_wrappers' => ['container'],
    ];

    $build['info_2'] = [
      '#markup' => $this->t('It will also find and replace all internal links with Linkit analogue, regardless if old link has redirect or not. Dependency to Linkit modules was done to make sure links in content are always point to a final destination. <br>'),
      '#theme_wrappers' => ['container'],
    ];

    $build['divider'] = [
      '#markup' => '<br>',
    ];

    $form = \Drupal::formBuilder()->getForm(SettingsForm::class);

    $build['settings_form'] = $form;

    $build['links'] = [
      'fix_node_links' => [
        '#type' => 'link',
        '#title' => $this->t('Find & Replace links with redirects in Nodes'),
        '#url' => Url::fromRoute('redirects_fixer.fix_node_links'),
        '#attributes' => ['class' => ['button']],
        '#access' => $this->moduleHandler()->moduleExists('node')
          && $this->config('redirects_fixer.settings')->get('site-domain'),
      ],
      'fix_blocks_links' => [
        '#type' => 'link',
        '#title' => $this->t('Find & Replace links with redirects in Blocks'),
        '#url' => Url::fromRoute('redirects_fixer.fix_block_links'),
        '#attributes' => ['class' => ['button']],
        '#access' => $this->moduleHandler()->moduleExists('block_content')
          && $this->config('redirects_fixer.settings')->get('site-domain'),
      ],
    ];

    $build['warning'] = [
      '#markup' => $this->t('<br> Warning! Clicking on this links will trigger replacement operation immediately! <br> It is recommended to run this script on staging first, then run SEO tool like Seo Screaming Frog to test carefully the results and site content after update!'),
      '#theme_wrappers' => ['container'],
      '#access' => (bool) $this->config('redirects_fixer.settings')->get('site-domain'),
    ];

    $name = 'redirect_fixer_log';
    $display_id = 'log_embed';
    $build['log'] = [
      '#type' => 'view',
      '#name' => $name,
      '#display_id' => $display_id,
    ];

    return $build;
  }

  /**
   * Fix links in nodes.
   */
  public function fixInNodes() {
    batch_set($this->batchManager->createBatchData('node'));
    return batch_process($this->reloadPage());
  }

  /**
   * Fix links in blocks.
   */
  public function fixInBlocks() {
    batch_set($this->batchManager->createBatchData('block_content'));
    return batch_process($this->reloadPage());
  }

  /**
   * Reload the previous page.
   */
  public function reloadPage() {
    $request = $this->requestStack->getCurrentRequest();
    if ($request->server->get('HTTP_REFERER')) {
      return $request->server->get('HTTP_REFERER');
    }
    else {
      return base_path();
    }
  }

}
