<?php

namespace Drupal\redirects_fixer;

use Drupal\Core\Entity\EntityInterface;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Client;
use GuzzleHttp\Exception\RequestException;
use GuzzleHttp\Promise\PromiseInterface;
use GuzzleHttp\Promise\Utils;

/**
 * Class LinksManager
 *
 * The LinksManager class is responsible for managing links and their destinations.
 */
class LinksManager {

  /**
   * The HTTP client to fetch the feed data with.
   *
   * @var \GuzzleHttp\ClientInterface
   */
  protected $httpClient;

  /**
   * Constructs a LinksManager object.
   */
  public function __construct(ClientInterface $http_client) {
    $this->httpClient = $http_client;
  }

  /**
   * Get urls destinations.
   */
  public function urlsDestination(array $urls): array {
    $client = new Client();
    $promises = [];
    foreach ($urls as $url) {
      $promises[$url] = $this->getFinalUrlPromise($client, $url);
    }

    $results = Utils::settle($promises)->wait();

    $final_urls = [];
    foreach ($results as $url => $result) {
      if ($result['state'] === 'fulfilled') {
        $final_urls[$url] = $result['value'];
      }
      else {
        $final_urls[$url] = $url;
      }
    }

    return $final_urls;
  }

  /**
   * Get request promise.
   */
  protected function getFinalUrlPromise(ClientInterface $client, $url): PromiseInterface {
    return $client->headAsync($url, [
      'allow_redirects' => [
        'max'             => 5,
        'strict'          => TRUE,
        'referer'         => TRUE,
        'protocols'       => ['http', 'https'],
        'track_redirects' => TRUE
      ],
      'http_errors' => FALSE,
      'timeout' => 10,
    ])->then(
      function ($response) use ($url) {
        $redirects = $response->getHeader('X-Guzzle-Redirect-History');
        $redirect = end($redirects);
        if ($redirect) {
          $statuses = $response->getHeader('X-Guzzle-Redirect-Status-History');
          if (end($statuses) == 301) {
            return $redirect;
          }
        }
        return $url;
      },
      function (RequestException $e) use ($url) {
        \Drupal::logger('redirects_fixer_getFinalUrlPromise')->error($url . ': ' . $e->getMessage());
        return $url;
      }
    );
  }

  /**
   * Update simple links element.
   */
  public function updateLink(\DOMElement $link_element, string $new_url): void {
    $to_remove_attr = [
      'data-entity-substitution',
      'data-entity-type',
      'data-entity-uuid',
    ];
    foreach ($to_remove_attr as $attr) {
      if ($link_element->hasAttribute($attr)) {
        $link_element->removeAttribute($attr);
      }
    }
    $link_element->setAttribute('href', $new_url);
  }

  /**
   * Update linkit links element.
   */
  public function updateLinkitLink(\DOMElement $link_element, string $new_url, EntityInterface $entity): void {
    $add_attr = [
      'data-entity-substitution' => ($entity->getEntityTypeId() === 'file' ? 'file' : 'canonical'),
      'data-entity-type' => $entity->getEntityTypeId(),
      'data-entity-uuid' => $entity->uuid(),
    ];
    foreach ($add_attr as $attr_name => $attr_value) {
      $link_element->setAttribute($attr_name, $attr_value);
    }
    $link_element->setAttribute('href', $new_url);
  }
}
